from ..base import NamedType
from ..base import Measurement


__all__ = ['Light']


@NamedType
class Light(Measurement):
	UVI: type
	Illuminance: type
	Irradiance: type
	RadiantFlux: type
	Lux: type
	Brightness: type


@NamedType
class UVI(Light):
	_max = 2


@NamedType
class Irradiance(Light):
	_unit = 'W/m²'

	@property
	def lux(self) -> 'Illuminance':
		"""Peter Michael, September 20, 2019, "A Conversion Guide: Solar Irradiance and Lux Illuminance ", IEEE Dataport, doi: https://dx.doi.org/10.21227/mxr7-p365."""
		return Illuminance(float(self) * 120)


@NamedType
class Illuminance(Light):
	_unit = 'lux'

	@property
	def wpm2(self) -> Irradiance:
		"""Peter Michael, September 20, 2019, "A Conversion Guide: Solar Irradiance and Lux Illuminance ", IEEE Dataport, doi: https://dx.doi.org/10.21227/mxr7-p365."""
		return Irradiance(float(self) / 120)


RadiantFlux = Irradiance
Lux = Illuminance
Brightness = Illuminance


Light.UVI = UVI
Light.Illuminance = Illuminance
Light.Irradiance = Irradiance

Light.RadiantFlux = Irradiance
Light.Lux = Illuminance
Light.Brightness = Illuminance



from ..config import config
from .. import UnitSystems

__all__ = ['NamedType', 'NamedSubType', 'UnitSystem', 'BaseUnit', 'Synonym', 'Tiny', 'Small', 'Medium', 'Large', 'Huge']

properties = config['UnitProperties']


def NamedType(cls):
	cls._type = cls
	cls._subTypes = {}
	cls.__isNamedType = True
	return cls


def NamedSubType(cls):
	parentCls = cls.__mro__[1]
	if hasattr(parentCls, 'genSubTypeName'):
		cls.__name__ = parentCls.genSubTypeName(cls)
	if not hasattr(parentCls, '_subTypes'):
		parentCls._subTypes = {}
	cls._subType = cls
	cls._siblingTypes = parentCls._subTypes
	return cls


def UnitSystem(cls):
	cls._unitSystem = cls
	system = cls.__name__.lower()
	unit = cls._type.__name__.lower()
	if system not in UnitSystems:
		UnitSystems[system] = {}
	if system != unit:
		cls.__name__ = f'{cls._type.__name__}({cls.__name__})'
	UnitSystems[system][unit] = cls
	return cls


def BaseUnit(cls):
	cls._unitSystem._baseUnit = cls
	cls._Scale._baseUnit = cls._Scale.Base
	return cls


def strToDict(string: str, cls: type) -> type:

	def parseString(item: str):
		key, value = item.split('=')
		# expectedTypes = {'max': int, 'precision': int, 'unitSpacer': stringToBool, 'shorten': stringToBool, 'thousandsSeparator': stringToBool, 'cardinal': stringToBool, 'degrees': stringToBool}
		if value.isnumeric():
			value = float(value)
			if value.is_integer():
				value = int(value)
		if value == 'True':
			value = True
		elif value == 'False':
			value = False
		return f'_{key}', value

	conf = [parseString(a) for a in [(y.strip(' ')) for y in string.split(',')]]
	for item in conf:
		setattr(cls, *item)

	return cls


def Tiny(cls):
	cls._size = 'tiny'
	return strToDict(properties['Tiny'], cls)


def Small(cls):
	cls._size = 'small'
	return strToDict(properties['Small'], cls)


def Medium(cls):
	cls._size = 'medium'
	return strToDict(properties['Medium'], cls)


def Large(cls):
	cls._size = 'large'
	return strToDict(properties['Large'], cls)


def Huge(cls):
	cls._size = 'huge'
	return strToDict(properties['Huge'], cls)


def Synonym(cls):
	cls.__name__ = cls.__mro__[1].__name__
	return cls



import os.path
from importlib import resources

import locale
import logging
from configparser import ConfigParser, SectionProxy
from pathlib import Path

from ..utils import convertString, setPropertiesFromConfig

log = logging.getLogger('WeatherUnitsConfig')


class Config(ConfigParser):
	configuredUnits: dict[str, list[type]] = {}
	locale = locale.getlocale()[0]
	__localUnits: SectionProxy

	def __init__(self, *args, path: str = None, locale: str = None, **kwargs):
		self.optionxform = str
		self.__localUnits = None
		path = path or os.environ.get('WU_CONFIG_PATH')
		super(Config, self).__init__(allow_no_value=True)

		if locale is not None:
			self.locale = locale
		if path is None:
			path = 'us.ini' if self.locale.lower().endswith('us') else 'si.ini'
			with resources.path(__package__, path) as path:
				self.path = path
		else:
			if isinstance(path, str):
				path = Path(path)
			self.path = path

		self.read(path, reloadModules=False)
		log.info(f'Loaded {path}')

	def __hash__(self):
		return hash(self.path)

	@property
	def customConfig(self):
		return self._customConfig

	def read(self, *args, reloadModules: bool = True, **kwargs):
		if args or kwargs:
			for arg in args:
				if os.path.isfile(arg):
					self.path = arg
					log.info(f'Loaded {arg}')
					super().read(arg, **kwargs)
					break
				else:
					log.error(f'{os.path.abspath(os.getcwd())}/{arg} is not a file')
			else:
				log.error(f'Unable to find config file')
		else:
			log.info(f'Loaded {self.path}')
		for k, v in self.configuredUnits.copy().items():
			setPropertiesFromConfig(v, self)

	def __getattr__(self, item):
		return self[item]

	@property
	def loc(self):
		return float(self['Location']['lat']), float(self['Location']['lon'])

	@property
	def lat(self):
		return float(self['Location']['lat'])

	@property
	def lon(self):
		return float(self['Location']['lon'])

	@property
	def unitProperties(self):
		return self['UnitProperties']

	@property
	def unitDefaults(self):
		return {f'_{prop}': convertString(value) for prop, value in config['UnitDefaults'].items()}

	@property
	def localUnits(self) -> SectionProxy:
		if self.__localUnits is None:
			if 'LocalUnits' in self:
				self.__localUnits = self['LocalUnits']
			elif 'Units' in self:
				self.__localUnits = self['Units']
			else:
				raise AttributeError(f'No local units defined in the config {self.path}')
		return self.__localUnits


config = Config()



from typing import Union

from enum import Enum

from .rate import DistanceOverTime
from ..base.Decorators import NamedType, NamedSubType
from ..length import Length
from ..time import Time

__all__ = ['Precipitation']


@NamedType
class Precipitation(Length):

	class Type(Enum):
		NONE = 0
		Rain = 1
		Hail = 2
		Sleet = 3
		Snow = 4

	def __new__(cls, value: Length, *args, **kwargs):
		value = value.__class__(value, *args, **kwargs)
		value.__dict__.update(Precipitation.__dict__)
		return value


@NamedType
class PrecipitationRate(DistanceOverTime, Precipitation):
	_numerator: Length
	_denominator: Time
	Daily: type
	Hourly: type
	Minutely: type

	def __init__(self, numerator: Length, denominator: int = 1, *args, **kwargs):
		if isinstance(denominator, int):
			denominator = Time.Hour(denominator)
		DistanceOverTime.__init__(self, numerator, denominator, *args, **kwargs)

	@property
	def fth(self):
		return Hourly(self._numerator.foot, self._denominator.hour)

	@property
	def inh(self):
		return Hourly(self._numerator.inch, self._denominator.hour)

	@property
	def ins(self):
		return PrecipitationRate(self._numerator.inch, self._denominator.second)

	@property
	def mmh(self):
		return Hourly(self._numerator.millimeter, self._denominator.hour)

	@property
	def mms(self):
		return PrecipitationRate(self._numerator.millimeter, self._denominator.second)

	@property
	def cmh(self):
		return Hourly(self._numerator.centimeter, self._denominator.hour)

	@property
	def mh(self):
		return Hourly(self._numerator.meter, self._denominator.hour)

	@property
	def daily(self):
		return Daily(self._numerator, self._denominator)

	@property
	def hourly(self):
		return Hourly(self._numerator, self._denominator)

	@property
	def minutely(self):
		return Minutely(self._numerator, self._denominator)

	@property
	def secondly(self):
		return PrecipitationRate(self._numerator, self._denominator)


@NamedSubType
class Daily(PrecipitationRate):
	_denominator: Time.Day

	def __init__(self, numerator: Length, denominator: Union[Time, int, float] = 1, *args, **kwargs):
		PrecipitationRate.__init__(self, numerator, denominator, *args, **kwargs)


@NamedSubType
class Hourly(PrecipitationRate):
	_denominator: Time.Hour

	def __init__(self, numerator: Length, denominator: Union[Time, int, float] = 1, *args, **kwargs):
		PrecipitationRate.__init__(self, numerator, denominator, *args, **kwargs)


@NamedSubType
class Minutely(PrecipitationRate):
	_denominator: Time.Minute

	def __init__(self, numerator: Length, denominator: Union[Time, int, float] = 1, *args, **kwargs):
		PrecipitationRate.__init__(self, numerator, denominator, *args, **kwargs)


@NamedSubType
class Secondly(PrecipitationRate):
	_denominator: Time.Second

	def __init__(self, numerator: Length, denominator: Union[Time, int, float] = 1, *args, **kwargs):
		PrecipitationRate.__init__(self, numerator, denominator, *args, **kwargs)


Precipitation.Rate = PrecipitationRate
Precipitation.Daily = Daily
Precipitation.Hourly = Hourly
Precipitation.Minutely = Minutely
Precipitation.Secondly = Secondly



from .temperature import Temperature
from .fahrenheit import Fahrenheit
from .celsius import Celsius
from .kelvin import Kelvin

__all__ = ['Temperature']

Temperature.Fahrenheit = Fahrenheit
Temperature.Celsius = Celsius
Temperature.Kelvin = Kelvin



from math import inf

from . import Temperature


class Fahrenheit(Temperature):
	_limits = (-459.6699, inf)
	_unit = 'f'

	def _celsius(self, delta: bool = False):
		if delta:
			return self / 1.8
		else:
			return (self - 32) / 1.8

	def _fahrenheit(self, delta: bool = False):
		return self

	def _kelvin(self):
		# (32°F − 32) × 1.8 + 273.15
		return (self - 32) / 1.8 + 273.15



from math import inf

from . import Temperature


class Kelvin(Temperature):
	_limits = (0, inf)
	_unit = 'k'

	def _kelvin(self):
		return self

	def _celsius(self, delta: bool = False):
		return self - 273.15

	def _fahrenheit(self, delta: bool = False):
		if delta:
			return (self - 273.15) * 1.8
		else:
			return ((self - 273.15) * 1.8) + 32



from math import inf

from . import Temperature


class Celsius(Temperature):
	_limits = (-273.15, inf)
	_unit = 'c'

	def _celsius(self, delta: bool = False):
		return self

	def _fahrenheit(self, delta: bool = False):
		if delta:
			return self * 1.8
		else:
			return (self * 1.8) + 32

	def _kelvin(self):
		return self + 273.15



from ..mass import Mass
from ..base import NamedType
from ..base.Measurement import DerivedMeasurement
from .volume import Volume

__all__ = ['Density']


@NamedType
class Density(DerivedMeasurement):
	_numerator: Mass
	_denominator: Volume



import logging
from typing import ClassVar, Optional, Set, Type, Union

from math import nan, isnan, inf
from decimal import Decimal

from ..utils import setPropertiesFromConfig
from ..config import config, Config

log = logging.getLogger('SmartFloat')

__all__ = ['SmartFloat']


class Meta(type):
	_derived: ClassVar[bool]

	# def __new__(cls, name, bases, dct):

	@property
	def isGenericType(cls):
		if '__isNamedType' in cls.__dict__:
			return cls.__dict__['__isNamedType']
		if '_unitSystem' in cls.__dict__:
			return cls.__dict__['_unitSystem'] is cls
		return False

	@property
	def isDerived(cls) -> bool:
		if not hasattr(cls, '_derived'):
			cls._derived = getattr(cls.__parent_class__, 'isDerived', None)
		return cls._derived

	@property
	def isScaling(cls) -> bool:
		return hasattr(cls, '_Scale')

	@property
	def __parent_class__(cls):
		return cls.__mro__[1]

	@property
	def subTypes(cls):
		subs: Set[Type['Measurement']] = set()
		if cls.isGenericType:
			for sub in cls.__subclasses__():
				if sub.isGenericType:
					subs.update(sub.subTypes)
				elif sub.isDerived == cls.isDerived:
					subs.add(sub)
		subs.discard(cls)
		return list(subs)

	@property
	def fixedUnits(cls) -> tuple[Optional[Type['Measurement']], Optional[Type['Measurement']]]:
		n: Type['Measurement'] = cls.__annotations__.get('_numerator', None)
		d: Type['Measurement'] = cls.__annotations__.get('_denominator', None)
		if n is not None:
			n = None if n.isGenericType else n
		if d is not None:
			d = None if d.isGenericType else d
		return n, d

	@property
	def limits(cls) -> tuple[Optional[float], Optional[float]]:
		if cls.isDerived:
			return cls.numeratorClass.limits
		return getattr(cls, '_limits', None) or getattr(cls.__parent_class__, 'limits', (0, inf))

	def __repr__(cls):
		if cls.isDerived:
			cls: 'DerivedMeasurement'
			if ((cls.numeratorClass.isGenericType or cls.__parent_class__.numeratorClass.isGenericType)
				and (cls.denominatorClass.isGenericType or cls.__parent_class__.denominatorClass.isGenericType)):
				return f'{cls.__name__}[{repr(cls.numeratorClass)}/{repr(cls.denominatorClass)}]'
			elif cls.numeratorClass.isGenericType or cls.__parent_class__.numeratorClass.isGenericType:
				return f'{cls.__name__}[{repr(cls.numeratorClass)}]'
			elif cls.denominatorClass.isGenericType or cls.__parent_class__.denominatorClass.isGenericType:
				return f'{cls.__name__}[{repr(cls.denominatorClass)}]'
			else:
				return f'{cls.__name__}'
		if cls.isGenericType:
			return f'<{cls.__name__}>'
		return cls.__name__


class SmartFloat(float, metaclass=Meta):
	_limits = 0, inf
	_config: Config
	_precision: int
	_max: int
	_unit: str
	_suffix: str
	_decorator: str
	_unitSpacer: bool
	_title: str
	_exp: int
	_slide: bool
	_showUnit: bool
	_shorten: bool
	_kSeparator: bool
	_combineUnitAndSuffix: bool
	_key: str
	_sizeHint: str
	_forcePrecision: bool
	_acceptedTypes: tuple = (float, int, Decimal)

	def __init_subclass__(cls, **kwargs):
		cls._acceptedTypes = (*kwargs.get('acceptedTypes', ()), *cls._acceptedTypes, cls)

		for prop in {key: value for key, value in config.unitDefaults.items() if not hasattr(cls, key)}.items():
			setattr(cls, *prop)
		setPropertiesFromConfig(cls, config)
		cls._config = config

	@classmethod
	def noneToNan(cls, value):
		if not isinstance(value, float):
			try:
				value = float(value)
			except ValueError:
				value = nan
			except TypeError:
				value = nan
		return value

	def __new__(cls, value):
		return float.__new__(cls, value)

	def __init__(self, value):
		float.__init__(value)

	@staticmethod
	def intFloatLength(value: float) -> tuple[int, int]:
		d, f = [len(i) for i in str(abs(float(value))).split('.')]
		return d, f

	def _string(self, hintString: bool = False, forceUnit: bool = None, multiplier: Union[int, float] = 1.0, asInt: bool = False) -> str:
		if self._shorten:
			c, valueFloat = 0, float(self * multiplier)
			numberLength = len(str(int(valueFloat)))
			while numberLength > 3 and numberLength >= self._max:
				c += 1
				valueFloat /= 1000
				numberLength = len(str(int(valueFloat)))
			suffix = ['', 'k', 'm', 'B'][c]
		else:
			valueFloat = float(self) * multiplier
			c = False
			suffix = ''
		if isnan(valueFloat):
			valueFloat = 0.0

		# TODO: Allow for precision to be overridden if number is scaled.  (10,110 becomes 10.11k instead of 10.1k)
		integerLength, floatingPointLength = self.intFloatLength(valueFloat)
		if floatingPointLength == 1 and valueFloat.is_integer():
			floatingPointLength = 0

		# Max amount of precision that can be displayed while keeping string under max length
		intAllowedPrecision = max(0, self._max - integerLength)
		precision = min(self._precision, integerLength)
		if hintString and floatingPointLength < precision and floatingPointLength < intAllowedPrecision:
			floatingPointLength = 1
		# Allow at least on level of precision if
		# Removed 1 if not decimal and c else decimal

		showUnit = self._showUnit if forceUnit is None else forceUnit
		floatingPointLength = max(intAllowedPrecision, precision) if self._forcePrecision else min(precision, intAllowedPrecision, floatingPointLength)
		formatStr = f"{{:{',' if self._kSeparator else ''}.{0 if asInt else floatingPointLength}f}}"
		valueString = formatStr.format(10 ** max(integerLength - 1, 0) if hintString else valueFloat)
		needsSpacer = (self._unitSpacer and showUnit) and self._unit or c
		spacer = " " if needsSpacer else ""
		unit = self.unit if showUnit else ''
		return f'{valueString}{suffix}{self._decorator}{spacer}{unit}'

	def __str__(self):
		return self._string()

	def __repr__(self):
		return self._string(forceUnit=True)

	def __bool__(self):
		return super().__bool__() or bool(self.unit)

	def __float__(self) -> float:
		return super().__float__()

	def __hash__(self):
		return hash(round(self, max(self._precision, 1)))

	@property
	def valueUnset(self) -> bool:
		return self == nan

	@property
	def withUnit(self):
		return self._string(forceUnit=True)

	@property
	def withoutUnit(self):
		return self._string(forceUnit=False)

	@property
	def unit(self) -> str:
		return self._unit

	@property
	def unitArray(self):
		return [self._unit]

	@property
	def suffix(self):
		return self._suffix

	@property
	def int(self):
		return int(self)

	@property
	def decoratedInt(self):
		return self._string(forceUnit=False, asInt=True)

	@property
	def name(self):
		return self.__class__.__name__

	@property
	def title(self):
		return self._title if self._title else self.name

	@title.setter
	def title(self, value):
		self._title = value

	@property
	def key(self) -> str:
		if self._key:
			return self._key
		else:
			t = self.title
			t = t.lower() if t.isupper() else t
			t = t.replace(' ', '')
			return f"{t[0].lower()}{t[1:]}"

	@key.setter
	def key(self, value: str):
		self._key = value

	@property
	def category(self):
		return self.__class__.__name__

	@property
	def showUnit(self):
		return self._showUnit

	@showUnit.setter
	def showUnit(self, value):
		self._showUnit = value

	@property
	def sizeHint(self) -> str:
		# length = len(string) - (0.5 * (string.count('.') + string.count(','))
		showUnit = not self._unitSpacer
		return self._string(hintString=True, forceUnit=showUnit).replace('1', '0') if self._sizeHint is None else self._sizeHint

	@property
	def decorator(self):
		return self._decorator

	@property
	def precision(self):
		return self._precision

	@precision.setter
	def precision(self, value):
		self._precision = value

	@property
	def max(self):
		return self._max

	@max.setter
	def max(self, value):
		self._max = value

	@property
	def shorten(self):
		return self._shorten

	@shorten.setter
	def shorten(self, value):
		self._shorten = value

	@property
	def unitSpacer(self):
		return self._unitSpacer

	@unitSpacer.setter
	def unitSpacer(self, value):
		self._unitSpacer = value



from math import inf

from functools import lru_cache
import logging
from datetime import datetime
from typing import Callable, ClassVar, Dict, List, Optional, Type

from .. import errors
from ..utils import HashSlice, loadUnitLocalization
from . import SmartFloat

__all__ = ['Measurement', 'DerivedMeasurement']

log = logging.getLogger('WeatherUnits')


class Measurement(SmartFloat):
	_derived = False
	_type: type = None
	_updateFunction: Optional[Callable] = None
	_timestamp: datetime
	_indoor: bool = False
	_calculated: bool = False
	_category: str = None
	_subTypes: ClassVar[Dict[str, Type['Measurement']]]

	def __new__(cls, value, title: str = None, key: str = None, timestamp: datetime = None, **kwargs):
		value = sorted((*cls.limits, float(value)))[1]
		return SmartFloat.__new__(cls, value)

	def __class_getitem__(cls, item):
		if (underItem := f'_{item}') in cls.__annotations__:
			return cls.__annotations__[underItem]
		if item in cls.__dict__:
			return cls.__dict__[item]
		if issubclass(item, Measurement):
			parentType = cls._type
			name = f'{parentType.__name__}[{item.__name__}]'
			newCls = type(name, (item, cls,), {})
			return newCls
		raise TypeError(f'{cls.__name__} does not support {item}')

	def __init__(self, value, title: str = None, key: str = None, timestamp: datetime = None, category: str = None):
		if isinstance(value, Measurement):
			if value._timestamp and timestamp is None:
				timestamp = value._timestamp
		if title:
			self._title = title
		if key:
			self._key = key
		self.category = category

		self._timestamp = timestamp

		SmartFloat.__init__(self, value)

	@property
	def category(self):
		if self._category is None:
			return self.__class__.__name__.lower()
		return self._category

	@category.setter
	def category(self, value):
		self._category = value

	@property
	def localize(self):
		if self.convertible:
			try:
				selector = loadUnitLocalization(type(self), self._config)
				selector = 'inch' if selector == 'in' else selector
				new = getattr(self, selector)
				new.title = self.title
				return new
			except AttributeError or KeyError as e:
				errors.Conversion.BadConversion("Unable to get localized type for {}".format(self.name), e)
		else:
			return self

	@property
	def convertible(self):
		try:
			return self._type.__name__.lower() in self._config.localUnits
		except AttributeError:
			return False

	@property
	def str(self):
		return str(self)

	@property
	def type(self) -> type:
		return self._type if self._type is not None else self.__class__.__mro__[0]

	@property
	def timestamp(self) -> datetime:
		return self._timestamp

	@property
	def updateFunction(self):
		return self._updateFunction

	@updateFunction.setter
	def updateFunction(self, function: callable):
		self._updateFunction = function

	@property
	def indoor(self) -> bool:
		return self._indoor

	@indoor.setter
	def indoor(self, value: bool):
		self._indoor = value

	@property
	def calculated(self):
		return self._calculated

	@calculated.setter
	def calculated(self, value):
		self._calculated = value

	def _convert(self, other):
		if isinstance(other, self.type):
			return self.__class__(other)
		if isinstance(other, (float, int)):
			return self.__class__(other)
		else:
			return other

	def transform(self, other, transformTo: type = None):
		if isinstance(other, type) and issubclass(other, Measurement):
			other = other(self)
		nonTransferred = ['_unit', '_suffix', '_scale', '_denominator', '_numerator']
		if self.type != other.type:
			log.warning(f'{self.withUnit} and {other.withUnit} are not identical types, this may cause issues')
		other.__dict__.update({key: value for key, value in self.__dict__.items() if key not in nonTransferred and value is not None})
		if other._updateFunction:
			other._updateFunction(other)
		return other

	def __eq__(self, other):
		if isinstance(other, SmartFloat):
			if not self.__class__ == other.__class__:
				try:
					other = self._convert(other)
				except errors.BadConversion:
					pass
			precision = min(self.precision, other.precision)
			return round(self, precision) == round(other, precision)
		elif isinstance(other, self._acceptedTypes):
			return self._convert(other) == self

	def __getitem__(self, item):
		try:
			return self.__getattribute__(item)
		except AttributeError:
			raise errors.Conversion.UnknownUnit(self, item)

	def __mul__(self, other):
		if isinstance(other, self.type):
			other = self._convert(other)
		if isinstance(other, self._acceptedTypes):
			other = self._convert(other)
		return self._convert(super().__mul__(other))

	def __add__(self, other):
		if isinstance(other, self.type):
			other = self._convert(other)
		elif isinstance(other, self._acceptedTypes):
			other = self._convert(other)
		return self._convert(super().__add__(other))

	def __radd__(self, other):
		if isinstance(other, self.type):
			other = self._convert(other)
		elif isinstance(other, self._acceptedTypes):
			other = self._convert(other)
		return self._convert(super().__radd__(other))

	def __sub__(self, other):
		if isinstance(other, self.type):
			other = self._convert(other)
		elif isinstance(other, self._acceptedTypes):
			other = self._convert(other)
		return self._convert(super().__sub__(other))

	def __rsub__(self, other):
		if isinstance(other, self.type):
			other = self._convert(other)
		elif isinstance(other, self._acceptedTypes):
			other = self._convert(other)
		return self._convert(super().__rsub__(other))

	def __pow__(self, power, modulo=None):
		if isinstance(power, self.type):
			power = self._convert(power)
		elif isinstance(power, self._acceptedTypes):
			power = self._convert(power)
		return self._convert(super().__pow__(power, modulo))

	def __truediv__(self, other):
		if isinstance(other, self.type):
			other = self._convert(other)
		elif isinstance(other, self._acceptedTypes):
			other = self._convert(other)

		if isinstance(other, self.type):
			value = super().__truediv__(self._convert(other))

		elif issubclass(self.__class__, DerivedMeasurement) and not isinstance(other, Measurement):
			return DerivedMeasurement(self, self.denominator.__class__(other))

		elif not isinstance(other, Measurement):
			value = super().__truediv__(float(other))

		elif issubclass(other.__class__, Measurement):
			return DerivedMeasurement(self, other)

		else:
			value = self / other

		return self._convert(value)

	def __gt__(self, other):
		if isinstance(other, self.type):
			other = self._convert(other)
		elif isinstance(other, self._acceptedTypes):
			other = self._convert(other)
		return super().__gt__(other)

	def __lt__(self, other):
		if isinstance(other, self.type):
			other = self._convert(other)
		elif isinstance(other, self._acceptedTypes):
			other = self._convert(other)
		return super().__lt__(other)

	def __ge__(self, other):
		if isinstance(other, self.type):
			other = self._convert(other)
		elif isinstance(other, self._acceptedTypes):
			other = self._convert(other)
		return super().__ge__(other)

	def __le__(self, other):
		if isinstance(other, self.type):
			other = self._convert(other)
		elif isinstance(other, self._acceptedTypes):
			other = self._convert(other)
		return super().__le__(other)

	def __hash__(self):
		return hash(round(self, max(self._precision, 1)))

	## Transform shortcuts ##

	def __or__(self, other):
		return self.transform(other)

	def __lshift__(self, other):
		return self.transform(other)


class DerivedMeasurement(Measurement):
	_derived = True
	_type: Type[Measurement] = Measurement
	_numerator: Measurement
	_denominator: Measurement
	numeratorClass: Type[Measurement]
	denominatorClass: Type[Measurement]

	def __new__(cls, numerator: Measurement, denominator: Measurement = 1, *args, **kwargs):
		if not cls.numeratorClass.isGenericType:
			numerator = cls.numeratorClass(numerator)
		if not cls.denominatorClass.isGenericType:
			denominator = cls.denominatorClass(denominator)
		value = float(numerator)/float(denominator)

		if isinstance(numerator, cls.numeratorClass) and isinstance(denominator, cls.denominatorClass):
			if cls.isGenericType:
				cls = cls[type(numerator):type(denominator)]
		elif items := [i for i in cls._subTypes if i['denominatorClass'] is type(denominator)]:
			if len(items) == 1:
				cls = items[0]
			else:
				items = [i for i in items if i['numeratorClass'] is type(numerator)]
				if len(items) == 1:
					cls = items[0]
				else:
					raise TypeError(f'{cls.__name__} does not support {numerator.__class__.__name__} as numerator')
		elif cls.isGenericType:
			cls = cls[type(numerator):type(denominator)]
		return Measurement.__new__(cls, value, *args, **kwargs)

	def __init_subclass__(cls, **kwargs):
		super().__init_subclass__(**kwargs)
		if 'numeratorClass' not in cls.__dict__:
			cls.numeratorClass = cls.__annotations__.get('_numerator', getattr(cls.__mro__[1], 'numeratorClass', cls.__mro__[1].__annotations__.get('_numerator', Measurement)))
		if 'denominatorClass' not in cls.__dict__:
			cls.denominatorClass = cls.__annotations__.get('_denominator', getattr(cls.__mro__[1], 'denominatorClass', cls.__mro__[1].__annotations__.get('_denominator', Measurement)))

	def __class_getitem__(cls, item):
		return cls.__getSlice__(cls, HashSlice(item))

	@lru_cache(maxsize=32)
	def __getSlice__(cls, item: HashSlice):
		fixedN, fixedD = cls.fixedUnits
		# if (underItem := f'_{item}') in cls.__annotations__:
		# 	return cls.__annotations__[underItem]
		if item in cls.__dict__:
			return cls.__dict__[item]
		if item.isSlice:
			sliceNType, sliceDType = fixedN or item.start, fixedD or item.stop
			nType, dType = cls.numeratorClass, cls.denominatorClass
			if sliceNType is not None and not isinstance(sliceNType, type):
				sliceNType = type(sliceNType)
			if sliceDType is not None and not isinstance(sliceDType, type):
				sliceDType = type(sliceDType)
			parentType = cls._type
			sameDType = {i for i in parentType.subTypes if i.denominatorClass is sliceDType}
			sameNType = {i for i in parentType.subTypes if i.numeratorClass is sliceNType}

			if exactMatch := sameDType & sameNType:
				return exactMatch.pop()

			if sameDType and (generics := {i for i in sameDType if i.numeratorClass.isGenericType}):
				if len(generics) == 1:
					generic = generics.pop()
					if not dType.isGenericType:
						return generic[sliceNType or nType]
					if generic.denominatorClass is sliceDType:
						return generic[sliceNType or nType]
			if sameNType and (generics := {i for i in sameNType if i.denominatorClass.isGenericType}):
				if len(generics) == 1:
					generic = generics.pop()
					if not nType.isGenericType:
						return generic[sliceDType or dType]
					if generic.numeratorClass is sliceNType:
						return generic[sliceDType or dType]

			name = f'{parentType.__name__}[{item.start.__name__}/{item.stop.__name__}]'
			return type(name, (parentType,), {'numeratorClass': sliceNType, 'denominatorClass': sliceDType})
		parentType = cls._subType
		item = item.start
		if issubclass(item, cls.numeratorClass):
			if not cls.denominatorClass.isGenericType:
				name = f'{cls.__name__}'
				newCls = type(name, (cls,), {'numeratorClass': item, 'denominatorClass': cls.denominatorClass})
				return newCls
			print(f'{cls.__name__} does not support {item.__name__} as numerator')

		elif issubclass(item, cls.denominatorClass):
			if not cls.numeratorClass.isGenericType:
				name = f'{cls.__name__}[{item.__name__}]'
				newCls = type(name, (cls,), {'denominatorClass': item})
				return newCls
			print(f'{cls.__name__} does not support {item.__name__} as denominator')
		return parentType[item]

	def __init__(self, numerator, denominator=None, *args, **kwargs):
		if isinstance(numerator, DerivedMeasurement) and denominator is None:
			self._numerator = numerator.numerator
			self._denominator = numerator.denominator
		elif denominator is None:
			self._numerator = numerator
			self._denominator = self.denominatorClass(1)
		else:
			self._numerator = numerator
			self._denominator = denominator
		if not isinstance(self._numerator, self.numeratorClass):
			self._numerator = self.numeratorClass(self._numerator)
		if not isinstance(self._denominator, self.denominatorClass):
			self._denominator = self.denominatorClass(self._denominator)
		Measurement.__init__(self, float(self._numerator)/float(self._denominator), *args, **kwargs)

	def __pow__(self, power, modulo=None):
		return self.__class__(super(Measurement, self).__pow__(power, modulo), self._denominator)

	def transform(self, other, transformTo: type = None):
		if isinstance(other, type) and issubclass(other, Measurement):
			other = other(self)
		nonTransferred = ['_unit', '_suffix', '_scale', '_denominator', '_numerator']
		if self.type != other.type:
			log.warning(f'{self.withUnit} and {other.withUnit} are not identical types, this may cause issues')
		other.__dict__.update({key: value for key, value in self.__dict__.items() if key not in nonTransferred and value is not None})
		if other._updateFunction:
			other._updateFunction(other)
		return other

	def _convert(self, other):
		if isinstance(other, (int, float)):
			return self.__class__(self.n.__class__(other), self.d.__class__(1))
		if isinstance(other, self.n.type):
			try:
				other = other[self.n.unit]
			except AttributeError:
				pass
			return self.__class__(other, self.d)
		else:
			return super(DerivedMeasurement, self)._convert(other)

	# TODO: Implement into child classes
	def _getUnit(self) -> List[str]:
		return self._config['LocalUnits'][str(self._type)].split('/')

	def _getUnitTypes(self):
		return self._numerator.type, self._denominator.type

	@property
	def type(self):
		return type(self)[type(self._numerator):type(self._denominator)]

	@property
	def localize(self):
		try:
			n, d = loadUnitLocalization(type(self), self._config).split('/')
			n = 'inch' if n == 'in' else n

			fixedN, fixedD = type(self).fixedUnits
			if n == '*' and fixedN:
				n = fixedN
			if d == '*' and fixedD:
				d = fixedD

			if isinstance(n, type):
				n = n._unit
			if isinstance(d, type):
				d = d._unit

			n = getattr(self._numerator, n)
			d = getattr(self._denominator, d)

			t = getattr(self, '_subType', self._type)
			tFN, tFD = t.fixedUnits
			if (tFN is not None and tFN != n) or (tFD is not None and tFD != d):
				t = t.__parent_class__

			newCls = t[type(n):type(d)]
			return newCls(n, d)
		except KeyError:
			log.error('Measurement {} was unable to localize from {}'.format(self.name, self.unit))
			return self

	@property
	def unit(self):
		if self._unit:
			return self._unit
		elif self._numerator.unit == 'mi' and self._denominator.unit == 'hr':
			return 'mph'
		elif issubclass(self.__class__, DerivedMeasurement):
			l = self.unitArray
			i = 0
			while i < len(l):
				power = 0
				while i + 1 < len(l) and l[i] == l[i + 1]:
					power += 1
					l.pop(i)
				if power:
					unicodePlace = 177 if power < 4 else 0x2074
					l[i] = f"{l[i]}{power + 177:c}"
				i += 1
			return '/'.join(l)
		else:
			return '{}/{}'.format(self._numerator.unit, self._denominator.unit)

	@property
	def unitArray(self) -> list[str]:
		return [*self._numerator.unitArray, *self._denominator.unitArray]

	@property
	def numerator(self) -> Measurement:
		return self._numerator

	n = numerator

	@property
	def denominator(self) -> Measurement:
		return self._denominator

	d = denominator

	@property
	def dArr(self):
		arr = []
		if isinstance(self.n, DerivedMeasurement):
			arr += [*self.n.dArr]
		if isinstance(self.d, DerivedMeasurement):
			arr += [*self.d.dArr]
		else:
			arr += [self.d.unit]
		return arr

	@property
	def nArr(self):
		arr = []
		if isinstance(self.d, DerivedMeasurement):
			arr += [*self.n.nArr]
		if isinstance(self.n, DerivedMeasurement):
			arr += [*self.n.nArr]
		else:
			arr += [self.n.unit]
		return arr



from ..length import Length
from ..time import Time
from ..base.Measurement import DerivedMeasurement
from ..base.Decorators import NamedType

from datetime import timedelta as _td

__all__ = ['DistanceOverTime']


@NamedType
class DistanceOverTime(DerivedMeasurement):
	_numerator: Length
	_denominator: Time

	def __new__(cls, numerator: Length, denominator: Time = 1):
		if isinstance(denominator, _td):
			denominator = Time.Second(denominator.total_seconds())
		return super().__new__(cls, numerator, denominator)

	def __init__(self, numerator: Length, denominator: Time = 1):
		if isinstance(denominator, _td):
			denominator = Time.Second(denominator.total_seconds())
		super().__init__(numerator, denominator)

	@property
	def fts(self):
		return DistanceOverTime(self._numerator.ft, self._denominator.s)

	@property
	def mih(self):
		converted = DistanceOverTime(self._numerator.mi, self._denominator.hr)
		converted._suffix = 'mph'
		return converted

	@property
	def inh(self):
		return DistanceOverTime(self._numerator.inch, self._denominator.hr)

	@property
	def ms(self):
		return DistanceOverTime(self._numerator.m, self._denominator.s)

	@property
	def mh(self):
		return DistanceOverTime(self._numerator.m, self._denominator.hr)

	@property
	def kmh(self):
		return DistanceOverTime(self._numerator.km, self._denominator.hr)

	@property
	def mmh(self):
		return DistanceOverTime(self._numerator.mm, self._denominator.hr)

	mph = mih
	fps = fts



from ..base import Scale, UnitSystem, BaseUnit, Large, Medium, Small, Tiny
from .length import Length


@UnitSystem
class Imperial(Length):
	class _Scale(Scale):
		Line = 1
		Inch = 12
		Foot = 12
		Yard = 3
		Mile = 1760
		Base = 'Foot'

	def _meter(self):
		return self.changeScale(self._Scale.Foot) * 0.3048


@Tiny
class Line(Imperial):
	_unit = 'ln'


@Small
class Inch(Imperial):
	_unit = 'in'


@Medium
@BaseUnit
class Foot(Imperial):
	_unit = 'ft'


@Medium
class Yard(Imperial):
	_unit = 'yd'


@Large
class Mile(Imperial):
	_unit = 'mi'
	_precision = 1




from ..base import Scale, UnitSystem, BaseUnit, Huge, Large, Small, Tiny
from . import Mass

__all__ = ['Metric', 'Microgram', 'Milligram', 'Gram', 'Kilogram', 'Tonne']


@UnitSystem
class Metric(Mass):

	class _Scale(Scale):
		Microgram = 1
		Milligram = 1000
		Gram = 1000
		Kilogram = 1000
		Tonne = 1000
		Base = 'Gram'

	def _pound(self):
		return self.changeScale(self._Scale.Kilogram) * 2.2046226218


@Tiny
class Microgram(Metric):
	_unit = 'μg'


@Tiny
class Milligram(Metric):
	_unit = 'mg'


@Small
@BaseUnit
class Gram(Metric):
	_unit = 'g'


@Large
class Kilogram(Metric):
	_unit = 'kg'


@Huge
class Tonne(Metric):
	_unit = 't'



class BadConversion(Exception):
	"""These two measurements can not be converted this way"""
	def __init__(self, first, second):
		self.message = f"{first} can not be converted to {second} this way"
		super().__init__(self.message)


class UnknownUnit(Exception):
	"""This measurement can not be converted to that unit"""
	def __init__(self, measurement, unit: str):
		self.message = f"{measurement} can not be converted to '{unit}'"
		super().__init__(self.message)



from datetime import timedelta
from typing import Any, Dict, Iterable, List, SupportsIndex, Type, Union

UnitSystems: Dict[str, Dict[str, Type]] = {}

from . import errors
from .config import config
from . import utils
from . import base
from . import temperature
from . import length
from . import mass
from . import time
from . import pressure
from .airQuality import *
from .derived import *
from .others import *
from .various import *

Temperature = temperature.Temperature
Length = length.Length
Mass = mass.Mass
Time = time.Time
Pressure = pressure.Pressure


class Trend(List[Measurement]):

	def __init__(self, value: Union[Measurement, List[Measurement], Type[Measurement]] = None, max: Union[int, Time, timedelta] = None):
		if isinstance(value, Measurement):
			value = [value]
		if isinstance(value, list):
			pass
		elif isinstance(value, type):
			value = [value(0)]
		metadata = value[0]
		self.title = metadata.title
		self.key = metadata.key
		self.category = metadata.category
		self._valueClass = type(metadata)
		self.max = max
		self.insertValue(value)
		super(Trend, self).__init__()

	def insertValue(self, value: Any):
		if isinstance(value, self._valueClass):
			self.append(value)
		elif isinstance(value, Iterable):
			for v in value:
				self.insertValue(v)
		else:
			raise TypeError("Value must be a list of Measurement or a Measurement")
		while self.listIsTooLong():
			self.pop(0)

	def listIsTooLong(self):
		if isinstance(self.max, int):
			return len(self) > self.max
		elif isinstance(self.max, Time):
			max = timedelta(seconds=self.max.seconds)
		start = self[0].timestamp
		end = self[-1].timestamp
		return end - start > self.max



from typing import Union

from ..length import Length
from ..time import Time
from .rate import DistanceOverTime
from ..base import NamedType, NamedSubType
from ..others import Direction

__all__ = ['Wind']


@NamedType
class Wind(DistanceOverTime):
	_numerator: Length
	_denominator: Time

	# TODO: Add support for setting both speed and direction
	__direction: Direction = None

	# def __init__(self, speed: DistanceOverTime = None, direction: Direction = None):
	# 	if direction is not None:
	# 		self.__direction = direction
	# 	if
	# 	super(Wind, self).__init__(DistanceOverTime)

	@property
	def direction(self):
		return self.__direction

	@direction.setter
	def direction(self, value):
		if not isinstance(value, Direction) and isinstance(value, (float, int)):
			value = Direction(value)
		self.__direction = value


@NamedSubType
class PerSecond(Wind):
	_denominator: Time.Second

	def __init__(self, numerator: Length, denominator: Union[Time, int, float] = 1, *args, **kwargs):
		denominator = Time.Second(denominator)
		Wind.__init__(self, numerator, denominator, *args, **kwargs)


@NamedSubType
class PerMinute(Wind):
	_denominator: Time.Minute

	def __init__(self, numerator: Length, denominator: Union[Time, int, float] = 1, *args, **kwargs):
		denominator = Time.Minute(denominator)
		Wind.__init__(self, numerator, denominator, *args, **kwargs)


@NamedSubType
class PerHour(Wind):
	_denominator: Time.Hour

	def __init__(self, numerator: Length, denominator: Union[Time, int, float] = 1, *args, **kwargs):
		denominator = Time.Hour(denominator)
		Wind.__init__(self, numerator, denominator, *args, **kwargs)




class SmartString(str):
	_title = ''
	_key = ''

	def __new__(cls, value: str, title: str = None, key: str = None):
		return str(value)

	def __init__(self, value: str, title: str = None, key: str = None):
		self._key = key
		self._title = title
		super(SmartString, self).__init__(value)

	@property
	def title(self):
		return self._title

	@property
	def key(self):
		return self._key

	def type(self):
		return type(self)

	def category(self):
		return self.__class__.__name__



import logging
from functools import lru_cache
from typing import Hashable, Type, Union

log = logging.getLogger(__name__)


@lru_cache(maxsize=64)
def loadUnitLocalization(measurement: Type['Measurement'], config):
	if not isinstance(measurement, type):
		measurement = measurement.__class__
	name = measurement.__name__
	possibleNames = [name.lower(), toCamelCase(name), name]

	if hasattr(measurement, '_type') and measurement._type is not None:
		possibleTypes = [measurement._type.__name__.lower(), toCamelCase(measurement._type.__name__), measurement._type.__name__]
		for classType in {*possibleNames, *possibleTypes}:
			if classType in config.localUnits:
				return config.localUnits[classType]
		else:
			log.warning(f'{measurement._type.__name__} has no type defined')


def toCamelCase(string, titleCase: bool = False) -> str:
	string = (string.lower() if string.isupper() else string)
	for char in ['-', ' ', '.', '_']:
		string.replace(char, '')
	return string[0].upper() if titleCase else string[0].lower() + string[1:]



def convertString(value: str) -> Union[int, float, str, bool]:
	if value is None:
		return value
	if value.isnumeric():
		value = float(value)
		if value.is_integer():
			value = int(value)
		return value
	if value == 'True':
		value = True
	elif value == 'False':
		value = False
	return value


def parseString(item: str):
	key, value = item.split('=')
	value = convertString(value)
	return f'_{key}', value


def strToDict(string: str, cls: type) -> type:
	conf = [parseString(a) for a in [(y.strip(' ')) for y in string.split(',')]]
	for item in conf:
		setattr(cls, *item)
	return cls


def setPropertiesFromConfig(cls, config) -> None:
	"""
	Modifies the class in place to add the properties defined in the config.
	:param cls: Class to modify
	:type cls: Type[Measurement]
	:param config: Config to use
	:type config: Config
	"""
	possibleNames = [cls.__name__.lower(), cls._unit, toCamelCase(cls.__name__), cls.__name__]
	if hasattr(cls, '_type') and cls._type is not None and not isinstance(cls._type, str):
		possibleTypes = [cls._type.__name__.lower(), toCamelCase(cls._type.__name__), cls._type.__name__]
		for classType in possibleTypes:
			try:
				cls = strToDict(config.unitProperties[classType], cls)
				config.configuredUnits[cls.__name__] = cls
				break
			except KeyError:
				pass
		else:
			log.debug(f'{cls.__name__} has no type defined')

	for name in possibleNames:
		try:
			cls = strToDict(config.unitProperties[name], cls)
			config.configuredUnits[name] = cls
			break
		except KeyError:
			pass


class HashSlice:
	__slots__ = ('start', 'stop', 'step')
	start: Union[Hashable, type]
	stop: Union[Hashable, type]
	step: Union[Hashable, type]

	def __init__(self, start: Union[Hashable, slice], stop: Hashable = None, step: Hashable = None):
		if isinstance(start, (slice, HashSlice)):
			stop = start.stop
			step = start.step
			start = start.start
		self.start = start
		self.stop = stop
		self.step = step

	def __iter__(self) -> iter:
		return iter((self.start, self.stop, self.step))

	def __repr__(self) -> str:
		return f"HashSlice[{self.start or ''}:{self.stop or ''}:{self.step or ''}]"

	def __str__(self) -> str:
		return f"[{self.start or ''}:{self.stop or ''}:{self.step or ''}]"

	def __eq__(self, other: object) -> bool:
		if hasattr(other, 'start') and hasattr(other, 'stop') and hasattr(other, 'step'):
			return self.start == other.start and self.stop == other.stop and self.step == other.step
		if self.stop is None and self.step is None:
			return self.start == other
		return False

	def __hash__(self) -> int:
		values = [self.start, self.stop, self.step]
		if self.step is None:
			values.pop()
		if self.stop is None and self.step is None:
			values.pop()
		if len(values) == 1:
			return hash(values[0])
		return hash(tuple(values))

	@property
	def isSlice(self) -> bool:
		return self.stop is not None



from .wind import *
from .rate import *
from .precipitation import *
from .volume import *
from .density import *
from .partsPer import *

__all__ = ['Volume', 'DistanceOverTime', 'Precipitation', 'Density']

DistanceOverTime.Wind = Wind



from typing import Union

from ..base.Decorators import NamedType
from ..base.Measurement import Measurement
from ..length import Length

__all__ = ['Volume']


@NamedType
class Volume(Measurement):

	CubicFoot: type
	CubicMeter: type

	_x: Length
	_y: Length
	_z: Length
	_unitClass = Length
	_cube: bool

	def __init__(self, x: Union[Length, int, float], y: Union[Length, int, float] = 0, z: Union[Length, int, float] = 0, cube: bool = True):
		if cube and not (z and y):
			x **= 1. / 3.
			y = x
			z = x
		self._cube = cube
		self._x: Length = self._unitClass(x)
		self._y: Length = self._unitClass(y)
		self._z: Length = self._unitClass(z)
		float.__init__(x * y * z)

	def __new__(cls, x: Length, y: Length = 1, z: Length = 1, cube: bool = False):
		if cube and not (z and y):
			y = x
			z = x
		value = x * y * z
		return float.__new__(cls, value)

	@property
	def width(self) -> Length:
		return self._x

	@property
	def length(self) -> Length:
		return self._y

	@property
	def depth(self) -> Length:
		return self._z

	@property
	def unit(self) -> str:
		return self._x.unit + "³"

	def scale(self, value: float):
		x = self._x * value
		y = self._y * value
		z = self._z * value
		return x, y, z


class CubicMeter(Volume):
	_x: Length.Meter
	_y: Length.Meter
	_z: Length.Meter
	_unitClass = Length.Meter

	@property
	def ft(self):
		x = self._x.ft
		y = self._y.ft
		z = self._z.ft
		return CubicFoot(x, y, z)

	@property
	def m(self):
		return self


class CubicFoot(Volume):
	_x: Length.Foot
	_y: Length.Foot
	_z: Length.Foot
	_unitClass = Length.Foot

	@property
	def ft(self):
		return self

	@property
	def m(self):
		x = self._x.m
		y = self._y.m
		z = self._z.m
		return CubicMeter(x, y, z)


Volume.CubicMeter = CubicMeter
Volume.CubicFoot = CubicFoot



from enum import Enum, EnumMeta
from typing import Optional, Union

from .. import errors
from ..base.Measurement import Measurement, DerivedMeasurement

__all__ = ['ScalingMeasurement', 'SystemVariant', 'Scale']


class AddressableEnum(EnumMeta):

	def __getitem__(self, indexOrName):
		if isinstance(indexOrName, str):
			return super().__getitem__(indexOrName)
		elif isinstance(indexOrName, int) and indexOrName < super().__len__():
			return self._list[indexOrName]
		elif isinstance(indexOrName, slice):
			indices = range(*indexOrName.indices(len(self._list)))
			return [self._list[i] for i in indices]

	@property
	def _list(self):
		return list(self)


class Indexer:
	# Todo: this could probably be moved inside or declared as a class variable of Scale

	i = 0
	lastClass: object = None

	@classmethod
	def get(cls, caller):
		cls.i, cls.lastClass = (0, caller) if not caller == cls.lastClass else (cls.i + 1, caller)
		return cls.i


class Scale(Enum, metaclass=AddressableEnum):

	def __new__(cls, *args):
		if isinstance(args[0], int):
			obj = object.__new__(cls)
			obj._value_ = Indexer.get(cls)
			obj._mul_ = args[0]
			obj.isVariant = False
		elif isinstance(args[0], float):
			obj = object.__new__(cls)
			obj._value_ = Indexer.get(cls)
			obj._mul_ = args[0]
			obj.isVariant = True
		elif isinstance(args[0], str):
			obj = object.__new__(cls)
			obj._value_ = getattr(cls, args[0]).index
			obj._mul_ = getattr(cls, args[0]).value
			obj.isVariant = False
		else:
			obj = None

		return obj

	def __str__(self):
		return str(self.value)

	# this makes sure that the description is read-only
	@property
	def index(self):
		if self.isVariant:
			return self.Base.index + 1
		return self._value_

	@property
	def value(self):
		return self._mul_

	def __repr__(self):
		return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, self._mul_)

	def __mul__(self, other):
		if isinstance(other, self.__class__):
			variant = self.isVariant or other.isVariant
			val = 1
			array = self.__class__[min(self.index, other.index) + 1: max(self.index, other.index) + 1]
			array.sort(key=lambda x: x.index)
			array = array[:-1] if variant else array
			for i in array:
				val *= i.value
			if variant:
				val *= self.value if self.isVariant else other.value
			return val

	def __gt__(self, other):
		if isinstance(other, self.__class__):
			return self.index > other.index
		else:
			return self.index > int(other)

	def __lt__(self, other):
		if isinstance(other, self.__class__):
			return self.index < other.index
		else:
			return self.index < int(other)

	def __eq__(self, other):
		if isinstance(other, self.__class__):
			return self.index == other.index
		else:
			return self.index == int(other)

	def __le__(self, other):
		if isinstance(other, self.__class__):
			return self.index <= other.index
		else:
			return self.index <= int(other)

	def __ge__(self, other):
		if isinstance(other, self.__class__):
			return self.index >= other.index
		else:
			return self.index >= int(other)


class ScalingMeasurement(Measurement):
	_baseUnit = None
	_unitSystem: type = None
	_Scale: Scale = None

	def __new__(cls, value, *args, **kwargs):
		if isinstance(value, DerivedMeasurement):
			return value
		value: Union[int, float, ScalingMeasurement, SystemVariant]
		if isinstance(value, ScalingMeasurement) and (not cls._baseUnit or not value.__class__._baseUnit):
			'''For this to work each unit class must have a _baseUnit defined for each scale'''
			raise errors.Unit.NoBaseUnitDefined(cls)
		system = issubclass(cls, ScalingMeasurement) and issubclass(value.__class__, ScalingMeasurement)
		siblings, cousins = [cls._baseUnit == value._baseUnit, cls._type == value._type] if system else (False, False)
		# variant = issubclass(cls, SystemVariant) or issubclass(value.__class__, SystemVariant)
		# If values are siblings change to sibling class with changeScale()
		if siblings:
			if isinstance(value, ScalingMeasurement) and not isinstance(value, value._baseUnit):
				value = value.toBaseUnit()
			return cls(value.__class__.changeScale(value, cls._Scale[cls.__name__]))

		# If values are cousins initiate with values base sibling causing a recursive call to __new__
		elif cousins:
			return cls.__new__(cls, cls._baseUnit(value.__getattribute__('_' + cls._baseUnit.__name__.lower())()))

		elif isinstance(value, Measurement):
			raise errors.Conversion.BadConversion(cls.__name__, value.__class__.__name__)

		else:
			return Measurement.__new__(cls, value, *args, **kwargs)

	def changeScale(self, newUnit: Scale) -> Optional[float]:
		if self._Scale:
			multiplier = self.scale * newUnit
			if self.scale > newUnit:
				return float(self) * multiplier
			else:
				return float(self) / multiplier
		else:
			return None

	def toBaseUnit(self) -> Measurement:
		return self._baseUnit(self.changeScale(self._Scale.Base))

	@property
	def scale(self):
		return self._Scale[self.name]

	@property
	def unitSystem(self):
		return self._unitSystem.__name__

	@property
	def isSystemVariant(self):
		return issubclass(self.__class__, SystemVariant)


class SystemVariant:
	_multiplier: float = 1.0



from ..base import NamedType, Synonym
from ..base import ScalingMeasurement
from ..base import BaseUnit, Huge, Large, Medium, Small, SystemVariant, Tiny, UnitSystem
from ..base.ScalingMeasurement import Scale
from enum import Enum


@UnitSystem
@NamedType
class Pressure(ScalingMeasurement):
	Pascal: type
	Decapascal: type
	Hectopascal: type
	Kilopascal: type
	Gigapascal: type
	Atmosphere: type
	TechnicalAtmosphere: type
	PoundsPerSquareInch: type
	MillimeterOfMercury: type
	InchOfMercury: type
	Bar: type
	Kilopascal: type
	Millibar: type
	PoundsPerSquareInch: type

	class Trend(Enum):
		Falling = -1
		Steady = 0
		Rising = 1

	class _Scale(Scale):
		Pascal = 1
		Decapascal = 10
		Hectopascal = 10
		Kilopascal = 10
		Megapascal = 1000
		Gigapascal = 1000
		Base = 'Pascal'
		Bar = 1000.
		Atmosphere = 101325.
		TechnicalAtmosphere = 98066.5
		PoundsPerSquareInch = 6894.757293168
		MillimeterOfMercury = 1 / 0.00750062
		InchOfMercury = 1 / 0.00029530

	@property
	def pascal(self):
		return Pascal(self)

	@property
	def decapascal(self):
		return Decapascal(self)

	@property
	def hectopascal(self):
		return Hectopascal(self)

	@property
	def kilopascal(self):
		return Kilopascal(self)

	@property
	def megapascal(self):
		return Megapascal(self)

	@property
	def gigapascal(self):
		return Gigapascal(self)

	@property
	def bar(self):
		return Bar(self)

	@property
	def millimeterOfMercury(self):
		return MillimeterOfMercury(self)

	@property
	def inchOfMercury(self):
		return InchOfMercury(self)

	@property
	def atmosphere(self):
		return Atmosphere(self)

	@property
	def technicalAtmosphere(self):
		return TechnicalAtmosphere(self)

	@property
	def poundsPerSquareInch(self):
		return PoundsPerSquareInch(self)

	@property
	def millibar(self):
		return Millibar(self)


@Tiny
@BaseUnit
class Pascal(Pressure):
	_unit = 'Pa'


@Small
class Decapascal(Pressure):
	_unit = 'daPa'


@Medium
class Hectopascal(Pressure):
	_unit = 'hPa'


@Large
class Kilopascal(Pressure):
	_unit = 'kPa'


@Huge
class Megapascal(Pressure):
	_unit = 'MPa'


@Huge
class Gigapascal(Pressure):
	_unit = 'GPa'


class Bar(Pascal, SystemVariant):
	_unit = 'bar'


class MillimeterOfMercury(Pascal, SystemVariant):
	_unit = "mmHg"


class InchOfMercury(Pascal, SystemVariant):
	_unit = 'inHg'


class Atmosphere(Pascal, SystemVariant):
	_unit = 'atm'


class TechnicalAtmosphere(Pascal, SystemVariant):
	_unit = 'at'


class PoundsPerSquareInch(Pascal, SystemVariant):
	_unit = 'psi'


@Synonym
class Millibar(Hectopascal):
	_unit = 'mBar'
	_max = 4


Pressure.Pascal = Pascal
Pressure.Decapascal = Decapascal
Pressure.Hectopascal = Hectopascal
Pressure.Kilopascal = Kilopascal
Pressure.Gigapascal = Gigapascal
Pressure.Atmosphere = Atmosphere
Pressure.Bar = Bar
Pressure.Millibar = Millibar
Pressure.InchOfMercury = InchOfMercury
Pressure.MillimeterOfMercury = MillimeterOfMercury
Pressure.Atmosphere = Atmosphere
Pressure.TechnicalAtmosphere = TechnicalAtmosphere
Pressure.PoundsPerSquareInch = PoundsPerSquareInch
Pressure.mbar = Pressure.mBar = Pressure.millibar
Pressure.psi = Pressure.poundsPerSquareInch
Pressure.at = Pressure.technicalAtmosphere
Pressure.atm = Pressure.atmosphere
Pressure.inHg = Pressure.inchOfMercury
Pressure.mmHg = Pressure.millimeterOfMercury
Pressure.gPa = Pressure.gigapascal
Pressure.kPa = Pressure.kilopascal
Pressure.hPa = Pressure.hectopascal
Pressure.p = Pressure.pascal
Pressure.dPa = Pressure.decapascal



from fractions import Fraction
from typing import Optional

from ..base import NamedType
from ..base import Measurement
from . import light

Light = light.Light

__all__ = ['Light', 'Angle', 'Direction', 'Humidity', 'Voltage', 'Strikes', 'Probability', 'Coverage', 'Percentage']


class Percentage(Measurement):
	_unit = ''
	_decorator = '%'
	_maxValue = 1.0
	_minValue = 0.0

	def __new__(cls, value, isPercentage: bool = None, limits: bool = True, *args, **kwargs):
		if isPercentage is None:
			if value > cls._maxValue:
				value /= 100
		elif isPercentage:
			value /= 100

		if limits:
			value = min(cls._maxValue, max(value, cls._minValue))

		return super(Percentage, cls).__new__(cls, value, *args, **kwargs)

	def _string(self, *args, **kwargs):
		kwargs['multiplier'] = 100
		return super(Percentage, self)._string(*args, **kwargs)

	def __int__(self) -> int:
		return int(float(self) * 100)

	def fromFloat(self, value: float) -> 'Percentage':
		return Percentage(value, isPercentage=False, limits=False)


@NamedType
class Humidity(Percentage):
	_decorator = '%'


@NamedType
class Probability(Measurement):
	_decorator = '%'
	_fraction: Optional[Fraction]
	_denominatorLimit: int = 10

	def __init__(self, *args, **kwargs):
		self._fraction: Optional[Fraction] = None
		super(Probability, self).__init__(*args, **kwargs)

	@property
	def fraction(self) -> Fraction:
		if self._fraction is None:
			self._fraction = Fraction(self).limit_denominator(self._denominatorLimit)
		return self._fraction


@NamedType
class Coverage(Percentage):
	_decorator = '%'


@NamedType
class Angle(Measurement):
	_precision = 0
	_decorator = 'º'
	_shorten = True


@NamedType
class Direction(Angle):
	_cardinal = True
	_decorator = 'º'
	_max = 3

	def _string(self, **kwargs) -> str:
		if self._cardinal and self._degrees:
			return f'{self.cardinal} ({super()._string(**kwargs)})'
		if self._cardinal:
			return f'{self.cardinal}'
		else:
			return f'{super()._string(**kwargs)}'

	@property
	def cardinal(self) -> 'Cardinal':
		return Cardinal(self)

	@property
	def angle(self) -> Angle:
		return Angle(self)

	@property
	def decorator(self) -> str:
		return self._decorator if not self._cardinal else ''

	@property
	def decoratedInt(self) -> str:
		return super()._string(forceUnit=False, asInt=True)


class Cardinal:
	__slots__ = '__direction'
	__dirsAbbrv = ['N', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE', 'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW']
	__dirsFull = ['North', 'North Northeast', 'Northeast', 'East Northeast', 'East', 'East Southeast', 'Southeast',
		'South Southeast', 'South', 'South Southwest', 'Southwest', 'West Southwest', 'West',
		'West Northwest', 'Northwest', 'North Northwest']
	__direction: Direction
	direction: Direction

	def __init__(self, direction: Direction):
		self.__direction = direction

	def __str__(self):
		if self.shorten:
			return self.abbrivated
		else:
			return self.full

	@property
	def direction(self) -> Direction:
		return self.__direction

	@property
	def __shortIndex(self):
		return round(self.direction/45)%8*2

	@property
	def __longIndex(self):
		return round(self.direction/22.5)%16

	@property
	def abbrivated(self):
		if self.direction.max > 2:
			return self.threeLetter
		return self.twoLetter

	@property
	def shortened(self):
		if self.direction.max < 5:
			return self.abbrivated
		return self.singleWord

	@property
	def twoLetter(self):
		return self.__dirsAbbrv[self.__shortIndex]

	@property
	def threeLetter(self):
		return self.__dirsAbbrv[self.__longIndex]

	@property
	def singleWord(self):
		return self.__dirsFull[self.__shortIndex]

	@property
	def text(self):
		if self.direction.max < 7:
			return self.singleWord
		return self.full

	@property
	def full(self):
		return self.__dirsFull[self.__longIndex]

	@property
	def shorten(self):
		return self.direction.shorten

	@shorten.setter
	def shorten(self, value: bool):
		self.direction.shorten = value


@NamedType
class Voltage(Measurement):
	_max = 3
	_precision = 2
	_unit = 'v'


@NamedType
class Strikes(Measurement):
	# TODO: Create discrete number class for this and similar measurements
	unit = 'strikes'
